/*
** Module   : ChuiWebSimulator.java
** Abstract : A WebSocket Chui Terminal simulator.
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- ----------------------Description--------------------------
** 001 MAG 20131113 Created initial version.
** 002 MAG 20140114 Implements WebSockets protocol and communication.
** 003 MAG 20140204 Send text attributes via WebSockets. Add a Watchdog Timer 
**                  guardian and quit command.                   
** 004 MAG 20140211 Implements COLOR.
*/

package com.goldencode.p2j.ui.client.chui.driver.web;

import java.io.*;
import java.nio.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.logging.*;
import java.util.regex.*;

import org.eclipse.jetty.websocket.api.*;
import org.eclipse.jetty.websocket.api.annotations.*;

import com.goldencode.p2j.main.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.swing.*;
import com.goldencode.p2j.util.*;

/**
 * Chui WebSocket terminal simulator.
 * Implements the WebSockets protocol to communicate with the JS terminal simulator.
 * On socket close terminate the client process.
 */
@SuppressWarnings("serial")
@WebSocket(maxIdleTime = 900000, maxTextMessageSize = 4096)
public class ChuiWebSimulator
extends ChuiSimulator
{  
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(ChuiWebSimulator.class.getName()); 
   
   /** Keyboard JSON pattern */
   private static final Pattern KEYBOARD = Pattern.compile("\\{\"c\":0,\"k\":([0-9]{1,4})\\}");
   
   /** Clear screen */
   private static final String INIT = "{\"c\":0}";
   
   /** Synchronization */
   private static final String SYNC = "{\"c\":1,\"t\":\"%s\"}"; 

   /** Set cursor position */
   private static final String SET_CURSOR_POS = "{\"c\":2,\"x\":\"%d\",\"y\":\"%d\"}"; 

   /** Show cursor */
   private static final String SET_CURSOR_ON = "{\"c\":3}"; 
   
   /** Hide cursor */
   private static final String SET_CURSOR_OFF = "{\"c\":4}"; 
   
   /** Play beep */
   private static final String BEEP = "{\"c\":5}"; 

   /** Quit */
   private static final String QUIT = "{\"c\":6, \"u\":\"%s\"}"; 

   /** Color palette */
   private static final String COLOR_PALETTE = "{\"c\":7, \"p\":%s}"; 

   /** Draw screen */
   private static final String DRAW_SCREEN = "{\"c\":8}";
   
   /**  Object representing the other end of this conversation. */
   private RemoteEndpoint remote;  
   
   /** Keyboard input queue */
   private Deque<Integer> keyBuffer = new ConcurrentLinkedDeque<>();
   
   /** Keep cursor status */
   private boolean cursorVisible = true;
   
   /** Time out value */
   private long timeout;
   
   /** Watchdog time out value */
   private long watchdogTimeout;

   /** Watchdog instance */
   private WatchdogTimer watchdog;
   
   /** Color palette JSON message */
   private String paletteMessage;
   
   /** Holds the last attributes message sent to JS code*/
   private volatile byte[] attributesCache;

   /** Holds the last colors message sent to JS code*/
   private volatile byte[] colorsCache;

   /** Holds the last screen message sent to JS code*/
   private volatile String screenCache;
   
   /**
    * Constructor.
    */
   public ChuiWebSimulator()
   {
      // build palette message
      paletteMessage = buildPaletteMessage();
      // Session timeout
      timeout = ClientBuilderOptions.getInstance().getWebSocketTimeout();
      // Watchdog timeout
      watchdogTimeout = ClientBuilderOptions.getInstance().getWatchdogTimeout();
      // Start watchdog timer
      startWatchdog(watchdogTimeout);
   }
   
   /**
    * This method is called when the remote peer open a web socket connection.
    *
    * @param   session
    *          An active link of communications with a Remote WebSocket Endpoint.
    *
    */
   @OnWebSocketConnect
   public void onConnect(Session session)
   {
      // Stop watchdog timer
      stopWatchdog();
      // Get RemoteEndpoint reference.
      this.remote = session.getRemote();
      // Set session timeout
      if (timeout > 0)
      {
         session.setIdleTimeout(timeout);
      }
      // synchronize tty
      initRemoteTerminal();
   }

   /**
    * This method is called when a text message is received.
    *
    * @param   message
    *          The text message.
    */
   @OnWebSocketMessage
   public void onMessage(String message) 
   {
      Matcher matcher = KEYBOARD.matcher(message);
      if (matcher.find())
      {
         String keystroke = matcher.group(1);
         keyBuffer.addLast(Integer.valueOf(keystroke));
      }                  
   }

   /**
    * This method is called when a binary message is received.
    *
    * @param   message
    *          Message content as an byte array.
    * @param   offset
    *          Message offset..
    * @param   length
    *          Message size.
    */
   @OnWebSocketMessage
   public void onMessage(byte[] message, int offset, int length)
   {
      // TODO implements binary protocol.
   }

   /**
    * This method is called when the connection is closed.
    * An END-ERROR KEY (F4) code is enqueued on key buffer queue.   
    *
    * @param   statusCode
    *          The status integer code.
    * @param   reason
    *          A string representing the reason.
    */
   @OnWebSocketClose
   public void onClose(int statusCode, String reason)
   {
      remote = null;
      // Start watchdog timer
      startWatchdog(watchdogTimeout);
      
      LOG.logp(Level.INFO,
               "ChuiWebSimulator.onClose",
               "",
               LogHelper.generate("Connection closed status=%d reason=%s", statusCode, reason));
   }
   
   /**
    * This method is called on communication errors.
    *
    * @param   session
    *          An active link of communications with a Remote WebSocket Endpoint.
    * @param   error
    *          A <code>Throwable</code> instance representing the error condition.
    */
   @OnWebSocketError
   public void onError(Session session, Throwable error)
   {
      session.close();
      remote = null;
      
      LOG.logp(Level.SEVERE,
               "ChuiWebSimulator.onError",
               "",
               LogHelper.generate("WebSocket communication error!"),
               error);      
   }

   /**
    * Start watchdog timer.
    * 
    * @param   watchdogTimeout
    *          Watchdog timeout value.
    */
   private void startWatchdog(long watchdogTimeout)
   {
      synchronized (keyBuffer)
      {
         if (watchdog == null)
         {
            watchdog = new WatchdogTimer(watchdogTimeout);
         }
      }      
   }

   /**
    * Stop watchdog timer.
    */
   private void stopWatchdog()
   {
      synchronized (keyBuffer)
      {
         if (watchdog != null)
         {
            watchdog.kill();
            // wait for thread exit
            try
            {
               watchdog.join(WatchdogTimer.WAIT_TO_DIE);
            }
            catch (InterruptedException e)
            {
               LOG.logp(Level.INFO,
                        "ChuiWebSimulator.stopWatchdog",
                        "",
                        LogHelper.generate("Watchdog interrupted!"),
                        e);      
            }
            finally
            {
               watchdog = null;
            }
         }
      }      
   }
   
   /**
    * Send a text message.
    * 
    * @param   message
    *          Text message.
    * 
    * @throws  IOException
    *          If cannot send the message.
    */   
   public void sendTextMessage(String message) 
   throws IOException
   {
      if (remote != null)
      {
         remote.sendString(message);
      }
   }

   /**
    * Send a binary message.
    * 
    * @param   message
    *          Binary message.
    * 
    * @throws  IOException
    *          If cannot send the message.
    */   
   public void sendBinaryMessage(byte[] message) 
   throws IOException
   {
      if (remote != null)
      {
         remote.sendBytes(ByteBuffer.wrap(message));
      }
   }
   
   /**
    * Whenever the main page is loaded on browser synchronize virtual terminal state.
    */
   private void initRemoteTerminal()
   {
      synchronized (keyBuffer)
      {
         // clear cache
         attributesCache = null;
         colorsCache = null;
         screenCache = null;
         // send color palette
         colorPalette();
         // synchronize
         triggerRepaint(true);
         setCursorStatus(cursorVisible);
         if (cursorVisible)
         {
            setCursor(getCursorX(), getCursorY());
         }
      }
   }
   
   /**
    * Read the next key from the FIFO key input buffer.  
    *
    * @return   The next key in the buffer or -1 if buffer is empty.
    */
   public int readKey()
   {
      return keyBuffer.isEmpty() ? -1 : keyBuffer.pollFirst();
   }
   
   /**
    * Overwrite the entire 2-dimensional output buffer with the space character
    * to be rendered with the <code>Color.NORMAL</code> color. Sets the cursor
    * Y and X positions to 0. Send INIT message down to JS client.
    */   
   @Override
   public void clear()
   {
      super.clear();
      try
      {
         sendTextMessage(INIT);
      }
      catch (IOException e)
      {
         LOG.logp(Level.SEVERE,
                  "ChuiWebSimulator.clear",
                  "",
                  LogHelper.generate("WebSocket communication error!"),
                  e);      
      }
   }

   /**
    * The beep() should be implemented by sending a command on the remote end point.
    * The command is part of the communication protocol.
    *
    */
   public void beep()
   {
      try
      {
         sendTextMessage(BEEP);
      }
      catch (IOException e)
      {
         LOG.logp(Level.SEVERE,
                  "ChuiWebSimulator.beep",
                  "",
                  LogHelper.generate("WebSocket communication error!"),
                  e);      
      }
   }
   
   /**
    *  Set the cursor visibility. Send SET_CURSOR_ON or SET_CURSOR_OFF message down to JS client.
    *
    * @param    on
    *           <code>true</code> to make the cursor visible (subsequent repaint
    *           operations will then cause the cursor to be drawn).
    *
    * @return   The previous cursor visibility state.
    */   
   @Override
   public boolean setCursorStatus(boolean on) 
   throws IllegalStateException
   {
      try
      {
         cursorVisible = on;
         sendTextMessage(on ? SET_CURSOR_ON : SET_CURSOR_OFF);
      }
      catch (IOException e)
      {
         LOG.logp(Level.SEVERE,
                  "ChuiWebSimulator.setCursorStatus",
                  "",
                  LogHelper.generate("WebSocket communication error!"),
                  e);      
      }
      // call super implementation
      return super.setCursorStatus(on);
   }

   /**
    * Position the cursor. Send SET_CURSOR_POS message down to JS client.
    *
    * @param    x
    *           New X cursor position.
    * @param    y
    *           New Y cursor position.
    *
    * @return  <code>true</code> if cursor positioning succeeded.
    */
   @Override
   public boolean setCursor(int x, int y)
   {
      try
      {
         // Convert to JSON and send
         String payload = String.format(SET_CURSOR_POS, y, x);
         sendTextMessage(payload);
      }
      catch (IOException e)
      {
         LOG.logp(Level.SEVERE,
                  "ChuiWebSimulator.setCursor",
                  "",
                  LogHelper.generate("WebSocket communication error!"),
                  e);      
      }
      // call super implementation
      return super.setCursor(x, y);
   }

   /**
    * Should be implemented by sending the screen status to the remote end point as a refresh
    * screen operation. The command is part of the communication protocol.
    * The colors and attributes messages are cached and performances are significantly improved.   
    */
   @Override
   public void triggerRepaint(boolean full)
   {
      super.triggerRepaint(full);
      
      TerminalOptions options = getTerminalOptions();
      int rows = options.rows;
      int columns = options.columns;
      
      Cell[][] data = getScreenData();
      
      // Attributes binary message
      // first byte is an id (0 = attributes)
      ByteArrayOutputStream attributes = new ByteArrayOutputStream(rows * columns);
      attributes.write(0); 

      ByteArrayOutputStream colors = new ByteArrayOutputStream(rows * columns);
      colors.write(1); 
      
      // Get screen characters
      StringBuffer screen = new StringBuffer();      
      for (int row = 0; row < rows; row++)
      {
         for (int col = 0; col < columns; col++)
         {
            Cell cell = data[row][col];
            // escape " character
            if (cell.character() == '"' || cell.character() == '\\')
            {
               screen.append('\\');
            }
            screen.append(cell.character());
            // attributes
            attributes.write(cell.color().attribute());
            // FG/BG colors
            int color = cell.color().color();
            byte bg = (byte) (color & 0x07);
            byte fg = (byte)((color >>> 12) & 0x70);
            colors.write(fg | bg);
         }
      }      
      
      synchronized (keyBuffer)
      {         
         try
         {         
            // Send FG/BG colors.
            sendCachedMessage(colors.toByteArray(), colorsCache);
            // Send attributes.
            sendCachedMessage(attributes.toByteArray(), attributesCache);
            // Send cached text message
            String screenCharacters = screen.toString();
            if(!screenCharacters.equals(screenCache))
            {
               screenCache = screenCharacters;
               sendTextMessage(String.format(SYNC, screenCharacters));
            }
            else
            {
               // Send draw screen command 
               sendTextMessage(DRAW_SCREEN);
            }
         }
         catch (IOException e)
         {
            LOG.logp(Level.SEVERE,
                     "ChuiWebSimulator.triggerRepaint",
                     "",
                     LogHelper.generate("WebSocket communication error!"),
                     e);      
         }
      }
   }

   /**
    * Send cached messages. The last send message is cached. The message is always compared with
    * the cached message. If the current message is equals to cached message the current message
    * is not send. If the current message is not equals with the cached message the cached
    * message is replaced by the current message and send back to JS client.
    * 
    * @param   currentMessage
    *          Current message.
    * @param   cachedMessage
    *          Cached message.
    *          
    * @throws  IOException
    *          Throw on send error.
    */
   private void sendCachedMessage(byte[] currentMessage, byte[] cachedMessage) 
   throws IOException
   {      
      if (!Arrays.equals(currentMessage, cachedMessage))
      {
         cachedMessage = Arrays.copyOf(currentMessage, currentMessage.length);
         sendBinaryMessage(currentMessage);
      }      
   }
   
   
   /**
    * Remote application is about to end.
    * Notify the JS client about that by sending a referrer address.
    * The client should be redirected to referrer address.
    * 
    * @param   url
    *          Referrer URL
    */
   public void quit(String url)
   {
      try
      {
         String payload = String.format(QUIT, url);
         sendTextMessage(payload);
      }
      catch (IOException e)
      {
         LOG.logp(Level.SEVERE,
                  "ChuiWebSimulator.quit",
                  "",
                  LogHelper.generate("WebSocket communication error!"),
                  e);      
      }
   }      
   
   /**
    * Build color palette message. This message is the same
    * over the user session so it's build and store on constructor.
    * The message is send whenever a WebSocket connection is open.
    *
    * @return  A string representing the JSON format of color palette.
    */
   private String buildPaletteMessage()
   {
      // build palette
      StringBuilder palette = new StringBuilder("[");      
      for (ColorPalette colorPalette : ColorPalette.values())
      {
         if (palette.length() != 1)
         {
            palette.append(",");
         }
         palette.append("\"").append(colorPalette.asWebColor()).append("\"");
      }      
      palette.append("]");      
      // JSON format
      return String.format(COLOR_PALETTE, palette.toString());
   }
   
   /**
    * Send color palette down to JS client.
    * This is done whenever a WebSocket connection is open.
    */
   private void colorPalette()
   {      
      try
      {
         // send palette
         sendTextMessage(paletteMessage);
      }
      catch (IOException e)
      {
         LOG.logp(Level.SEVERE,
                  "ChuiWebSimulator.colorPalette",
                  "",
                  LogHelper.generate("WebSocket communication error!"),
                  e);      
      }
   }
}